"""Content classes for rendering objects and markdown to HTML."""

import base64
from abc import ABC, abstractmethod
from io import BytesIO, StringIO
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from uuid import uuid4

import markdown as md

from esparto import _INSTALLED_MODULES
from esparto._options import options
from esparto._publish import nb_display
from esparto._typing import Child
from esparto._utils import responsive_svg_mpl

if "PIL" in _INSTALLED_MODULES:
    import PIL.Image as Img  # type: ignore
    from PIL.Image import Image as PILImage

if "pandas" in _INSTALLED_MODULES:
    from pandas import DataFrame  # type: ignore

if "matplotlib" in _INSTALLED_MODULES:
    from matplotlib.figure import Figure as MplFigure  # type: ignore

if "bokeh" in _INSTALLED_MODULES:
    from bokeh.embed import components  # type: ignore
    from bokeh.models.layouts import LayoutDOM as BokehObject  # type: ignore

if "plotly" in _INSTALLED_MODULES:
    from plotly.graph_objs._figure import Figure as PlotlyFigure  # type: ignore
    from plotly.io import to_html as plotly_to_html  # type: ignore

if TYPE_CHECKING:
    from esparto._layout import Row

T = TypeVar("T", bound="Content")


class Content(ABC):
    """Template for Content elements.

    Attributes:
        content (Any): Item to be included in the page - should match the encompassing Content class.

    """

    content: Any
    _dependencies: Set[str]

    @abstractmethod
    def to_html(self, **kwargs: bool) -> str:
        """Convert content to HTML string.

        Returns:
            str: HTML string.

        """
        raise NotImplementedError

    def display(self) -> None:
        """Display rendered content in a Jupyter Notebook cell."""
        nb_display(self)

    def __add__(self, other: Child) -> "Row":
        from esparto._layout import Row

        return Row(children=[self, other])

    def __iter__(self) -> Iterator["Content"]:
        return iter([self])

    def __len__(self) -> int:
        return len(list(self.content))

    def _repr_html_(self) -> None:
        nb_display(self)

    def __str__(self) -> str:
        return getattr(self, "title", None) or self.__class__.__name__

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            if hasattr(self.content, "__iter__") and hasattr(other.content, "__iter__"):
                return all(x == y for x, y in zip(self.content, other.content))
            return bool(self.content == other.content)
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)


class RawHTML(Content):
    """Raw HTML content.

    Args:
        html (str): HTML string.

    """

    _dependencies: Set[Any] = set("")
    content: str

    def __init__(self, html: str) -> None:

        if not isinstance(html, str):
            raise TypeError(r"HTML must be str")

        self.content = html

    def to_html(self, **kwargs: bool) -> str:
        return self.content


class Markdown(Content):
    """Markdown text content.

    Args:
        text (str): Markdown text to be added to document.

    """

    _dependencies = {"bootstrap"}

    def __init__(self, text: str) -> None:

        if not isinstance(text, str):
            raise TypeError(r"text must be str")

        self.content: str = text

    def to_html(self, **kwargs: bool) -> str:
        html = md.markdown(self.content, extensions=["extra", "smarty"])
        html = f"{html}\n"
        html = f"<div class='es-markdown'>\n{html}\n</div>"
        return html


class Image(Content):
    """Image content.

    Can be read from a filepath, PIL.Image object, or from bytes.

    Note:
        Requires optional `Pillow` library.

    Only one of `scale`, `set_width`, or `set_height` should be used.
    If more than one is populated, the values will be prioritised in the order:
         `set_width` -> `set_height` -> `scale`

    Args:
        image (str, PIL.Image, BytesIO): Image data.
        caption (str): Image caption (default = None)
        alt_text (str): Alternative text. (default = None)
        scale (float): Scale image proportionately, must be > 0 and <= 1. (default = None)
        set_width (int): Set width in pixels. (default = None)
        set_height (int): Set height in pixels. (default = None)

    """

    _dependencies = {"bootstrap"}

    def __init__(
        self,
        image: Union[str, Path, "PILImage", BytesIO],
        caption: Optional[str] = "",
        alt_text: Optional[str] = "Image",
        scale: Optional[float] = None,
        set_width: Optional[int] = None,
        set_height: Optional[int] = None,
    ):
        if "PIL" not in _INSTALLED_MODULES:
            raise ModuleNotFoundError("Install `Pillow` for image content support.")

        if not isinstance(image, (str, Path, PILImage, BytesIO)):
            raise TypeError(r"`image` must be one of {str, Path, PIL.Image, BytesIO}")

        self.content = image
        self.alt_text = alt_text
        self.caption = caption
        self._scale = scale
        self._width = set_width
        self._height = set_height

    def set_width(self, width: int) -> None:
        """Set width of image.

        Args:
            width (int): New width in pixels.

        """
        self._width = width

    def set_height(self, height: int) -> None:
        """Set height of image.

        Args:
            height (int): New height in pixels.

        """
        self._height = height

    def rescale(self, scale: float) -> None:
        """Resize the image by a scaling factor prior to rendering.

        Note:
            Images can be scaled down only.

        Args:
            scale (float): Scaling ratio.

        """
        self._scale = scale

    def to_html(self, **kwargs: bool) -> str:
        if isinstance(self.content, PILImage):
            image = self.content
        else:
            image = Img.open(self.content)

        if self._width or self._height or self._scale:
            image = _rescale_image(image, self._width, self._height, self._scale)

        image_encoded = _image_to_base64(image)

        width, _ = image.size

        html = (
            "<figure class='es-figure'>"
            "<img class='img-fluid figure-img rounded es-image' "
            f"style='width: min({int(width)}px, 100%);' "
            f"alt='{self.alt_text}' "
            f"src='data:image/png;base64,{image_encoded}'>"
        )

        if self.caption:
            html += f"<figcaption class='figure-caption'>{self.caption}</figcaption>"

        html += "</figure>"

        return html


class DataFramePd(Content):
    """Pandas DataFrame to be converted to table.

    Args:
        df (pd.DataFrame): A Pandas DataFrame
        index (bool): If True, render the DataFrame index. (default = True)
        col_space (str, int): Minimum column width in CSS units. Use int for pixels. (default = 0)

    Attributes:
        css_classes (List[str]): CSS classes applied to the HTML output.

    """

    _dependencies = {"bootstrap"}

    def __init__(
        self, df: "DataFrame", index: bool = True, col_space: Union[int, str] = 0
    ):

        if not isinstance(df, DataFrame):
            raise TypeError(r"df must be Pandas DataFrame")

        self.content: "DataFrame" = df
        self.index = index
        self.col_space = col_space
        self.css_classes = [
            "table",
            "table-hover",
        ]

    def to_html(self, **kwargs: bool) -> str:
        html: str = self.content.to_html(
            index=self.index,
            border=0,
            col_space=self.col_space,
            classes=self.css_classes,
        )
        html = f"<div class='table-responsive es-table'>{html}</div>"
        return html


class FigureMpl(Content):
    """Matplotlib figure.

    Args:
        figure (plt.Figure): A Matplotlib figure.
        output_format (str): 'svg' or 'png'. (default = None)
        pdf_figsize (tuple, float): Set figure size for PDF output. (default = None)
            Accepts a tuple of (height, width) or a float to use as scale factor.

    """

    _dependencies = {"bootstrap"}

    def __init__(
        self,
        figure: "MplFigure",
        output_format: Optional[str] = None,
        pdf_figsize: Optional[Union[Tuple[int, int], float]] = None,
    ) -> None:

        if not isinstance(figure, MplFigure):
            raise TypeError(r"figure must be a Matplotlib Figure")

        self.content: MplFigure = figure
        self.output_format = output_format or options.matplotlib.html_output_format
        self.pdf_figsize = pdf_figsize or options.matplotlib.pdf_figsize

        self._original_figsize = figure.get_size_inches()

    def to_html(self, **kwargs: bool) -> str:

        if kwargs.get("notebook_mode"):
            output_format = options.matplotlib.notebook_format
        else:
            output_format = self.output_format

        if kwargs.get("pdf_mode") and self.pdf_figsize:
            if isinstance(self.pdf_figsize, float):
                figsize = self.pdf_figsize * self._original_figsize
            else:
                figsize = self.pdf_figsize
            self.content.set_size_inches(*figsize)

        if output_format == "svg":

            string_buffer = StringIO()
            self.content.savefig(string_buffer, format="svg")
            string_buffer.seek(0)
            xml = string_buffer.read()

            dpi = 96
            fig_width, fig_height = (
                int(val * dpi) for val in self.content.get_size_inches()
            )

            if kwargs.get("pdf_mode"):
                xml = responsive_svg_mpl(
                    xml, width=int(fig_width), height=int(fig_height)
                )
                temp_file = Path(options._pdf_temp_dir) / f"{uuid4()}.svg"
                temp_file.write_text(xml)
                inner = (
                    "<object type='image/svg+xml' width='100%' height='100%' "
                    f"data='{temp_file.name}'></object>\n"
                )
            else:
                xml = responsive_svg_mpl(xml)
                inner = xml

            html = (
                f"<div class='es-matplotlib-figure' style='width: min({fig_width}px, 100%);'>"
                f"{inner}\n</div>\n"
            )

            # Reset figsize in case it was changed
            self.content.set_size_inches(*self._original_figsize)

            return html

        # If not svg:
        bytes_buffer = BytesIO()
        self.content.savefig(bytes_buffer, format="png")
        bytes_buffer.seek(0)
        return Image(bytes_buffer).to_html()


class FigureBokeh(Content):
    """Bokeh object to be rendered as an interactive plot.

    Args:
        figure (bokeh.layouts.LayoutDOM): A Bokeh object.
        layout_attributes (dict): Attributes set on `figure`. (default = None)

    """

    _dependencies = {"bokeh"}

    def __init__(
        self,
        figure: "BokehObject",
        layout_attributes: Optional[Dict[Any, Any]] = None,
    ):
        if not issubclass(type(figure), BokehObject):
            raise TypeError(r"figure must be a Bokeh object")

        self.content: BokehObject = figure
        self.layout_attributes = layout_attributes or options.bokeh.layout_attributes

    def to_html(self, **kwargs: bool) -> str:

        if self.layout_attributes:
            for key, value in self.layout_attributes.items():
                setattr(self.content, key, value)

        # Bokeh to PDF is experimental and untested
        if kwargs.get("pdf_mode"):  # pragma: no cover
            from bokeh.io import export_svg  # type: ignore

            temp_file = Path(options._pdf_temp_dir) / f"{uuid4()}.svg"
            export_svg(self.content, filename=str(temp_file))
            html = f"<img src='{temp_file.name}' width='100%' height='auto'>\n"
            return html

        html, js = components(self.content)

        # Remove outer <div> tag so we can give our own attributes
        html = _remove_outer_div(html)

        fig_width = self.content.properties_with_values().get("width", 1000)

        return (
            "<div class='es-bokeh-figure' "
            f"style='width: min({fig_width}px, 100%);'>"
            f"\n{html}\n{js}\n</div>"
        )


class FigurePlotly(Content):
    """Plotly figure to be rendered as an interactive plot.

    Args:
        figure (plotly.graph_objs._figure.Figure): A Plotly figure.
        layout_args (dict): Args passed to `figure.update_layout()`. (default = None)

    """

    _dependencies = {"plotly"}

    def __init__(
        self,
        figure: "PlotlyFigure",
        layout_args: Optional[Dict[Any, Any]] = None,
    ):

        if not isinstance(figure, PlotlyFigure):
            raise TypeError(r"figure must be a Plotly Figure")

        self.layout_args = layout_args or options.plotly.layout_args

        self.content: PlotlyFigure = figure
        self._original_layout = figure.layout

    def to_html(self, **kwargs: bool) -> str:

        if self.layout_args:
            self.content.update_layout(**self.layout_args)

        # Default width is 700, default height is 450
        fig_width: int = getattr(self.content, "width", 700)

        if kwargs.get("pdf_mode"):
            temp_file = Path(options._pdf_temp_dir) / f"{uuid4()}.svg"
            self.content.write_image(str(temp_file))
            inner = f"<img src='{temp_file.name}' width='100%' height='auto'>"

        else:
            inner = plotly_to_html(
                self.content, include_plotlyjs=False, full_html=False
            )
            # Remove outer <div> tag so we can give our own attributes.
            inner = _remove_outer_div(inner)

        html = f"<div class='es-plotly-figure' style='width: min({fig_width}px, 100%);'>{inner}\n</div>"

        # Reset layout in case it was changed
        self.content.update_layout(self._original_layout)

        return html


def _remove_outer_div(html: str) -> str:
    """Remove outer <div> tags."""
    html = html.replace("<div>", "", 1)
    html = "".join(html.rsplit("</div>", 1))
    return html


def _image_to_base64(image: "PILImage") -> str:
    """
    Convert an image from PIL to base64 representation.

    Args:
        image (PIL.Image):

    Returns:
        str: Image encoded as a base64 utf-8 string.

    """
    buffer = BytesIO()
    image.save(buffer, format="png")
    image_encoded = base64.b64encode(buffer.getvalue()).decode("utf-8")
    return image_encoded


def _rescale_image(
    image: "PILImage",
    width: Optional[int] = None,
    height: Optional[int] = None,
    scale: Optional[float] = None,
) -> "PILImage":
    """Rescale image by width in px, height in px, or ratio."""
    image = image.copy()
    new_size = _rescale_dims(image.size, width, height, scale)
    image.thumbnail(new_size)
    return image


def _rescale_dims(
    size: Tuple[int, int],
    width: Optional[int] = None,
    height: Optional[int] = None,
    scale: Optional[float] = None,
) -> Tuple[int, int]:
    """Rescale dimensions by width in px, height in px, or ratio."""
    if width:
        ratio = width / size[0]
    elif height:
        ratio = height / size[1]
    elif scale:
        ratio = scale
    else:
        raise ValueError("One of {'width', 'height', scale'} must be supplied")

    if ratio > 1:
        raise ValueError("Target size must be less than original size")

    new_size = (int(size[0] * ratio), int(size[1] * ratio))
    return new_size



"""Esparto configuration options."""

import collections.abc
import copy
import pprint
import traceback
from dataclasses import dataclass, field
from pathlib import Path
from types import TracebackType
from typing import Any, Dict, Mapping, Optional, Tuple, Type, Union

import yaml  # type: ignore

from esparto import _MODULE_PATH
from esparto._utils import public_dict


class ConfigMixin(object):
    _options_source: str

    def _to_dict(self) -> Dict[str, Any]:
        return public_dict(self.__dict__)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        string = f"{pprint.pformat(self._to_dict())}"
        if hasattr(self, "_options_source"):
            string += (
                f"\nSource: {self._options_source}" if self._options_source else ""
            )
        return string


@dataclass(repr=False)
class MatplotlibOptions(ConfigMixin):
    """Options for Matplotlib output.

    Attributes:
        html_output_format (str):
            How plots are rendered in HTML: 'png' or 'svg'.
        notebook_format (str):
            How plots are rendered in Jupyter Notebooks: 'png' or 'svg'.
        pdf_figsize (tuple or int):
            Specify size of Matplotlib figures in PDF output.
            An integer tuple can be passed as: (height, width).
            A float can be passed as a scaling factor.

    """

    html_output_format: str = "svg"
    notebook_format: str = "svg"
    pdf_figsize: Optional[Union[Tuple[int, int], float]] = 1.0


@dataclass(repr=False)
class PlotlyOptions(ConfigMixin):
    """Options for Plotly output.

    Attributes:
        layout_args (dict):
            Arguments passed to `figure.update_layout()` at rendering time.

    """

    layout_args: Dict[str, Any] = field(default_factory=lambda: {})


@dataclass(repr=False)
class BokehOptions(ConfigMixin):
    """Options for Bokeh output.

    Attributes:
        layout_attributes (dict):
            Bokeh layout object attributes to set at rendering time.

    """

    layout_attributes: Dict[str, Any] = field(
        default_factory=lambda: {"sizing_mode": "scale_width"}
    )


@dataclass(repr=False)
class OutputOptions(ConfigMixin):
    """Options for configuring page rendering and output.

    Config options will automatically be loaded if a yaml file is found at
    either `./esparto-config.yaml` or `~/esparto-data/esparto-config.yaml`.

    Attributes:
        dependency_source (str):
            How dependencies should be provisioned: 'cdn' or 'inline'.
        bootstrap_cdn (str):
            Link to Bootstrap CDN. Used if dependency source is 'cdn'.
            Alternative links are available from `esparto.bootstrap_cdn_themes`.
        bootstrap_css (str):
            Path to Bootstrap CSS file. Used if dependency source is 'inline'.
        esparto_css (str):
            Path to additional CSS file with esparto specific styles.
        jinja_template (str):
            Path to Jinja HTML page template.

        matplotlib: Additional config options for Matplotlib.
        plotly: Additional config options for Plotly.
        bokeh: Additional config options for Bokeh.

    """

    dependency_source: str = "cdn"
    bootstrap_cdn: str = (
        "<link rel='stylesheet' "
        "href='https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/5.1.3/css/bootstrap.min.css' "
        "integrity='sha512-GQGU0fMMi238uA+a/bdWJfpUGKUkBdgfFdgBm72SUQ6BeyWjoY/ton0tEjH+OSH9iP4Dfh+7HM0I9f5eR0L/4w==' "
        "crossorigin='anonymous' referrerpolicy='no-referrer'>"
    )
    bootstrap_css: str = str(_MODULE_PATH / "resources/css/bootstrap.min.css")
    esparto_css: str = str(_MODULE_PATH / "resources/css/esparto.css")
    jinja_template: str = str(_MODULE_PATH / "resources/jinja/base.html.jinja")

    matplotlib: MatplotlibOptions = MatplotlibOptions()
    bokeh: BokehOptions = BokehOptions()
    plotly: PlotlyOptions = PlotlyOptions()

    _pdf_temp_dir: str = ".pdf-temp"

    _options_source: str = ""

    def save(self, path: Union[str, Path] = "./esparto-config.yaml") -> None:
        """Save config to yaml file at `path`."""
        path = Path(path)
        self_copy = copy.copy(self)
        del self_copy._options_source
        yaml_str = yaml.dump(self_copy)
        path.write_text(yaml_str)

    def load(self, path: Union[str, Path]) -> None:
        """Load config from yaml file at `path`."""
        path = Path(path)
        yaml_str = path.read_text()
        self = yaml.unsafe_load(yaml_str)
        self._options_source = str(path)

    def _autoload(self) -> None:
        config_paths = [
            Path("./esparto-config.yaml"),
            Path.home() / "esparto-data/esparto-config.yaml",
        ]
        for p in config_paths:
            if p.is_file():
                self.load(p)
                print("esparto config loaded from:", p)
                self._options_source = str(p)
                break


options = OutputOptions()


def update_recursive(
    source_dict: Dict[Any, Any], update_map: Mapping[Any, Any]
) -> Dict[Any, Any]:
    """Recursively update nested dictionaries.
    https://stackoverflow.com/a/3233356/8065696
    """
    for k, v in update_map.items():
        if isinstance(v, collections.abc.Mapping):
            source_dict[k] = update_recursive(source_dict.get(k, {}), v)
        else:
            source_dict[k] = v
    return source_dict


class OptionsContext:
    def __init__(self, page_options: OutputOptions):
        self.page_options = page_options
        self.default_options = copy.copy(options)

    def __enter__(self) -> None:
        update_recursive(options.__dict__, self.page_options.__dict__)

    def __exit__(
        self, exc_type: Type[BaseException], exc_value: BaseException, tb: TracebackType
    ) -> None:
        if exc_type is not None:  # pragma: no cover
            traceback.print_exception(exc_type, exc_value, tb)
        update_recursive(options.__dict__, self.default_options.__dict__)


def resolve_config_option(config_option: str, value: Optional[str]) -> Any:
    if value is None:
        return getattr(options, config_option)
    else:
        return value



import re
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional

if TYPE_CHECKING:
    from esparto._typing import Child


def get_index_where(
    condition: Callable[..., bool], iterable: Iterable[Any]
) -> List[int]:
    """Return index values where `condition` is `True`."""
    return [idx for idx, item in enumerate(iterable) if condition(item)]


def get_matching_titles(title: str, children: List["Child"]) -> List[int]:
    """Return child items with matching title."""
    return get_index_where(lambda x: bool(getattr(x, "title", None) == title), children)


def clean_attr_name(attr_name: str) -> str:
    """Remove invalid characters from the attribute name."""
    if not attr_name:
        return ""

    # Remove leading and trailing spaces
    attr_name = attr_name.strip().replace(" ", "_").lower()

    # Remove invalid characters
    attr_name = re.sub("[^0-9a-zA-Z_]", "", attr_name)

    # Remove leading characters until we find a letter or underscore
    attr_name = re.sub("^[^a-zA-Z_]+", "", attr_name)

    return attr_name


def clean_iterator(iterator: Iterable[Any]) -> Iterable[Any]:
    # Convert any non-list iterators to lists
    iterator = (
        list(iterator) if isinstance(iterator, (list, tuple, set)) else [iterator]
    )
    # Un-nest any nested lists of children
    if len(list(iterator)) == 1 and isinstance(list(iterator)[0], (list, tuple, set)):
        iterator = list(iterator)[0]
    return iterator


def public_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    """Remove keys starting with '_' from dict."""
    return {
        k: v for k, v in d.items() if not (isinstance(k, str) and k.startswith("_"))
    }


def render_html(
    tag: str,
    classes: List[str],
    styles: Dict[str, str],
    children: str,
    identifier: Optional[str] = None,
) -> str:
    """Render HTML from provided attributes."""
    class_str = " ".join(classes) if classes else ""
    class_str = f"class='{class_str}'" if classes else ""

    style_str = "; ".join((f"{key}: {value}" for key, value in styles.items()))
    style_str = f"style='{style_str}'" if styles else ""

    id_str = f"id='{identifier}'" if identifier else ""

    rendered = " ".join((f"<{tag} {id_str} {class_str} {style_str}>").split())
    rendered += f"\n  {children}\n</{tag}>"

    return rendered


def responsive_svg_mpl(
    source: str, width: Optional[int] = None, height: Optional[int] = None
) -> str:
    """Make SVG element responsive."""

    regex_w = r"width=\S*"
    regex_h = r"height=\S*"

    width_ = f"width='{width}px'" if width else ""
    height_ = f"height='{height}px'" if height else ""

    source = re.sub(regex_w, width_, source, count=1)
    source = re.sub(regex_h, height_, source, count=1)

    # Preserve aspect ratio of SVG
    old_str = r"<svg"
    new_str = '<svg class="svg-content-mpl" preserveAspectRatio="xMinYMin meet" '
    source = source.replace(old_str, new_str, 1)

    return source



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['esparto']

package_data = \
{'': ['*'], 'esparto': ['resources/css/*', 'resources/jinja/*']}

install_requires = \
['beautifulsoup4>=4.7', 'jinja2>=2.10.1', 'markdown>=3.1', 'pyyaml>=5.1']

extras_require = \
{':python_version < "3.7"': ['dataclasses'],
 'extras': ['Pillow>=7.0.0', 'weasyprint>=51']}

setup_kwargs = {
    'name': 'esparto',
    'version': '3.0.1',
    'description': 'Data driven report builder for the PyData ecosystem.',
    'long_description': 'esparto\n=======\n\n[![image](https://img.shields.io/pypi/v/esparto.svg)](https://pypi.python.org/pypi/esparto)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/esparto.svg)](https://pypi.python.org/pypi/esparto/)\n![Build Status](https://github.com/domvwt/esparto/actions/workflows/lint-and-test.yml/badge.svg)\n[![codecov](https://codecov.io/gh/domvwt/esparto/branch/main/graph/badge.svg?token=35J8NZCUYC)](https://codecov.io/gh/domvwt/esparto)\n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=domvwt_esparto&metric=alert_status)](https://sonarcloud.io/dashboard?id=domvwt_esparto)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/esparto)\n\nIntroduction\n------------\n\n**esparto** is a Python library for building data driven reports with content\nfrom popular analytics packages. The project takes a straightforward approach\nto document design; with a focus on usability, portability, and extensiblity.\n\nCreating a report is as simple as instantiating a Page object and \'adding\' content\nin the form of DataFrames, plots, and markdown text. Documents can be built interactively\nin a notebook environment, and the results shared as a self-contained HTML\npage or PDF file.\n\nFurther customisation of the output is possible by passing a CSS stylesheet,\nchanging the [Jinja](Jinja) template, or declaring additional element styles within\nthe code. The responsive [Bootstrap](Bootstrap) grid ensures documents adapt to\nany viewing device.\n\nBasic Usage\n-----------\n\n```python\nimport esparto as es\n\n# Do some analysis\npandas_dataframe = ...\nplotly_figure = ...\n\n# Create a Page object\npage = es.Page(title="My Report")\n\n# Add content\npage["Data Analysis"]["Plot"] = plotly_figure\npage["Data Analysis"]["Data"] = pandas_dataframe\n\n# Save to HTML or PDF\npage.save_html("my-report.html")\npage.save_pdf("my-report.pdf")\n\n```\n\nMain Features\n-------------\n\n- Interactive document design with Jupyter Notebooks\n- Share as self-contained HTML or PDF\n- Customise with CSS and Jinja\n- Responsive Bootstrap grid layout\n- Content adaptors for:\n    - [Markdown][Markdown]\n    - [Images][Pillow]\n    - [Pandas DataFrames][Pandas]\n    - [Matplotlib][Matplotlib]\n    - [Bokeh][Bokeh]\n    - [Plotly][Plotly]\n\nInstallation\n------------\n\n**esparto** is available from [PyPI][PyPI] and [Conda][Conda]:\n\n```bash\npip install esparto\n```\n\n```bash\nconda install esparto -c conda-forge\n```\n\nDependencies\n------------\n\n- [python](https://python.org/) >= 3.6\n- [jinja2](https://palletsprojects.com/p/jinja/)\n- [markdown](https://python-markdown.github.io/)\n- [BeautifulSoup](https://www.crummy.com/software/BeautifulSoup/)\n- [PyYAML](https://pyyaml.org/)\n\n#### Optional\n\n- [Pillow](https://python-pillow.org/) *(for image content)*\n- [weasyprint](https://weasyprint.org/) *(for PDF output)*\n\nLicense\n-------\n\n[MIT](https://opensource.org/licenses/MIT)\n\nDocumentation\n-------------\n\nDocumentation and examples are available at\n[domvwt.github.io/esparto/](https://domvwt.github.io/esparto/).\n\nContributions, Issues, and Requests\n-----------------------------------\n\nFeedback and contributions are welcome - please raise an issue or pull request\non [GitHub][GitHub].\n\nExamples\n--------\n\nIris Report - [Webpage](https://domvwt.github.io/esparto/examples/iris-report.html) |\n[PDF](https://domvwt.github.io/esparto/examples/iris-report.pdf) | [Notebook](https://github.com/domvwt/esparto/blob/main/docs/examples/iris-report.ipynb)\n\n<br>\n\n<p width=100%>\n<img width=100%  src="https://github.com/domvwt/esparto/blob/1857f1d7411f12c37c96f8f5d60ff7012071851f/docs/images/iris-report-compressed.png?raw=true" alt="example page" style="border-radius:0.5%;">\n</p>\n\n<!-- * Links -->\n[PyPI]: https://pypi.org/project/esparto/\n[Conda]: https://anaconda.org/conda-forge/esparto\n[Bootstrap]: https://getbootstrap.com/\n[Jinja]: https://jinja.palletsprojects.com/\n[Markdown]: https://www.markdownguide.org/\n[Pillow]: https://python-pillow.org/\n[Pandas]: https://pandas.pydata.org/\n[Matplotlib]: https://matplotlib.org/\n[Bokeh]: https://bokeh.org/\n[Plotly]: https://plotly.com/\n[GitHub]: https://github.com/domvwt/esparto\n',
    'author': 'Dominic Thorn',
    'author_email': 'dominic.thorn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://domvwt.github.io/esparto',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



from types import SimpleNamespace


class BootstrapThemeCDNPicker(SimpleNamespace):
    """
    Links to Bootstrap theme CDNs that can be passed to `es.options.bootrap_cdn`.

    These themes are provided for free by Bootswatch.
    Visit https://bootswatch.com/4/ for example pages.

    """

    CERULEAN: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/cerulean/bootstrap.min.css" integrity="sha384-3fdgwJw17Bi87e1QQ4fsLn4rUFqWw//KU0g8TvV6quvahISRewev6/EocKNuJmEw" crossorigin="anonymous">'
    COSMO: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/cosmo/bootstrap.min.css" integrity="sha384-5QFXyVb+lrCzdN228VS3HmzpiE7ZVwLQtkt+0d9W43LQMzz4HBnnqvVxKg6O+04d" crossorigin="anonymous">'
    CYBORG: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/cyborg/bootstrap.min.css" integrity="sha384-nEnU7Ae+3lD52AK+RGNzgieBWMnEfgTbRHIwEvp1XXPdqdO6uLTd/NwXbzboqjc2" crossorigin="anonymous">'
    DARKLY: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/darkly/bootstrap.min.css" integrity="sha384-nNK9n28pDUDDgIiIqZ/MiyO3F4/9vsMtReZK39klb/MtkZI3/LtjSjlmyVPS3KdN" crossorigin="anonymous">'
    FLATLY: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/flatly/bootstrap.min.css" integrity="sha384-qF/QmIAj5ZaYFAeQcrQ6bfVMAh4zZlrGwTPY7T/M+iTTLJqJBJjwwnsE5Y0mV7QK" crossorigin="anonymous">'
    JOURNAL: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/journal/bootstrap.min.css" integrity="sha384-QDSPDoVOoSWz2ypaRUidLmLYl4RyoBWI44iA5agn6jHegBxZkNqgm2eHb6yZ5bYs" crossorigin="anonymous">'
    LITERA: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/litera/bootstrap.min.css" integrity="sha384-enpDwFISL6M3ZGZ50Tjo8m65q06uLVnyvkFO3rsoW0UC15ATBFz3QEhr3hmxpYsn" crossorigin="anonymous">'
    LUMEN: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/lumen/bootstrap.min.css" integrity="sha384-GzaBcW6yPIfhF+6VpKMjxbTx6tvR/yRd/yJub90CqoIn2Tz4rRXlSpTFYMKHCifX" crossorigin="anonymous">'
    LUX: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/lux/bootstrap.min.css" integrity="sha384-9+PGKSqjRdkeAU7Eu4nkJU8RFaH8ace8HGXnkiKMP9I9Te0GJ4/km3L1Z8tXigpG" crossorigin="anonymous">'
    MATERIA: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/materia/bootstrap.min.css" integrity="sha384-B4morbeopVCSpzeC1c4nyV0d0cqvlSAfyXVfrPJa25im5p+yEN/YmhlgQP/OyMZD" crossorigin="anonymous">'
    MINTY: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/minty/bootstrap.min.css" integrity="sha384-H4X+4tKc7b8s4GoMrylmy2ssQYpDHoqzPa9aKXbDwPoPUA3Ra8PA5dGzijN+ePnH" crossorigin="anonymous">'
    PULSE: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/pulse/bootstrap.min.css" integrity="sha384-L7+YG8QLqGvxQGffJ6utDKFwmGwtLcCjtwvonVZR/Ba2VzhpMwBz51GaXnUsuYbj" crossorigin="anonymous">'
    SANDSTONE: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/sandstone/bootstrap.min.css" integrity="sha384-zEpdAL7W11eTKeoBJK1g79kgl9qjP7g84KfK3AZsuonx38n8ad+f5ZgXtoSDxPOh" crossorigin="anonymous">'
    SIMPLEX: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/simplex/bootstrap.min.css" integrity="sha384-FYrl2Nk72fpV6+l3Bymt1zZhnQFK75ipDqPXK0sOR0f/zeOSZ45/tKlsKucQyjSp" crossorigin="anonymous">'
    SKETCHY: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/sketchy/bootstrap.min.css" integrity="sha384-RxqHG2ilm4r6aFRpGmBbGTjsqwfqHOKy1ArsMhHusnRO47jcGqpIQqlQK/kmGy9R" crossorigin="anonymous">'
    SLATE: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/slate/bootstrap.min.css" integrity="sha384-8iuq0iaMHpnH2vSyvZMSIqQuUnQA7QM+f6srIdlgBrTSEyd//AWNMyEaSF2yPzNQ" crossorigin="anonymous">'
    SOLAR: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/solar/bootstrap.min.css" integrity="sha384-NCwXci5f5ZqlDw+m7FwZSAwboa0svoPPylIW3Nf+GBDsyVum+yArYnaFLE9UDzLd" crossorigin="anonymous">'
    SPACELAB: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/spacelab/bootstrap.min.css" integrity="sha384-F1AY0h4TrtJ8OCUQYOzhcFzUTxSOxuaaJ4BeagvyQL8N9mE4hrXjdDsNx249NpEc" crossorigin="anonymous">'
    SUPERHERO: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/superhero/bootstrap.min.css" integrity="sha384-HnTY+mLT0stQlOwD3wcAzSVAZbrBp141qwfR4WfTqVQKSgmcgzk+oP0ieIyrxiFO" crossorigin="anonymous">'
    UNITED: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/united/bootstrap.min.css" integrity="sha384-JW3PJkbqVWtBhuV/gsuyVVt3m/ecRJjwXC3gCXlTzZZV+zIEEl6AnryAriT7GWYm" crossorigin="anonymous">'
    YETI: str = '<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootswatch@4.5.2/dist/yeti/bootstrap.min.css" integrity="sha384-mLBxp+1RMvmQmXOjBzRjqqr0dP9VHU2tb3FK6VB0fJN/AOu7/y+CAeYeWJZ4b3ii" crossorigin="anonymous">'

    def __repr__(self) -> str:
        return str(self) + f"\n{type(self)}"

    def __str__(self) -> str:
        themes = "\n".join([x for x in dir(self) if not x.startswith("_")])
        string = f"Available themes:\n{themes}"
        return string


bootstrap_cdn_themes = BootstrapThemeCDNPicker()
